
/*
 *  project     ontic
 *  Author      TriX
 *  Description main function for the codegenerator
 */

#include <iostream.h>
#include <stdlib.h>
#include <stdio.h>
#include <dir.h>

#include "readasm.h"

char *IndependendLibraries[] = {"ontics", "memmgt", "cmain",
                                "list", "searchkw", NULL};

int main (int argc, char *argv[])
{
   int Libs = 0, done, i;
   ReadASM  *Input[32];
   struct   ffblk f;
   ofstream Lib1 ("lib1.asm"), Lib2 ("lib2.asm"), PubLib ("publib.asm");
   int      L1c, L2c, pc;
   int      L1_Funcs, L2_Funcs, Publib_Funcs, TotalFuncs;
   int      ReSolve, Doors;
   char     TMP[256];

   cout << "[mxm] ontic code arranger (codegen)\n"
        << "by MXM Projects   http://mxm.ticalc.org/\n"
        << "by MXM-TriX       mxmtrix@gmx.net\n";

   if (argc != 2)
   {
      cout << "\nCalling convention\n"
              "ontic -[f|d]\n"
              "\t-f generate Fargo program\n"
              "\t-d generate DoorsOS program\n\n";
      return 0;
   } // if

   if (strcmp (argv[1], "-d") == 0 || strcmp (argv[1], "-D") == 0)
   {
      Doors = 1;
      cout << "Generating a DoorsOS program for ti92+ / ti89\n";
   }
   else if (strcmp (argv[1], "-f") == 0 || strcmp (argv[1], "-f") == 0)
   {
      Doors = 0;
      cout << "Generating a Fargo program for ti92\n";
   }
   else
   {
      cout << "\nCalling convention\n"
              "ontic -[f|d]\n"
              "\t-f generate Fargo program\n"
              "\t-d generate DoorsOS program\n\n";
      return 0;
   } // if

   cout << "Parsing assemblerfiles\n";

   done = findfirst ("*.s", &f, FA_ARCH);
   while (!done)
   {
      Input[Libs]     = new ReadASM (f.ff_name, IndependendLibraries, Doors);
      Input[Libs + 1] = NULL;
      Libs++;
      done = findnext (&f);
   } // while

   cout << "Solving references\n";
   ReSolve = 1;
   for (ReSolve = 0; ReSolve < 4; ReSolve++)
   {
      for (i = 0; i < Libs; i++)
         Input[i]->Solve (Input);

      for (i = 0; i < Libs; i++)
      {
         TotalFuncs   += Input[i]->GetFunctions ();
         L1_Funcs     += Input[i]->MovedFunctions (LIB1);
         L2_Funcs     += Input[i]->MovedFunctions (LIB2);
         Publib_Funcs += Input[i]->MovedFunctions (PUBLIB);
      } // for

   }

   cout << "Generating new sourcecode\n";
   Lib1   << "; [auto code]\n\txdef _library\n\n\tinclude \"onticsf.h\"\n";
   Lib2   << "; [auto code]\n\txdef _library\n\n\tinclude \"onticsf.h\"\n";
   PubLib << "; [auto code]\n\txdef _library\n\n\tinclude \"onticsf.h\"\n";
   L1c = L2c = pc = 0;
   for (i = 0; i < Libs; i++)
   {
      Input[i]->WriteDefinitions (Lib1,   LIB1,   L1c);
      Input[i]->WriteDefinitions (Lib2,   LIB2,   L2c);
      Input[i]->WriteDefinitions (PubLib, PUBLIB, pc);
   } // for

   sprintf (TMP, "%04X", pc);
   PubLib << "\txdef publib@" << TMP << "\n\n__start:\n";

   // Write function positions (librarypositions)
   for (i = 0; i < Libs; i++)
   {
      Input[i]->WritePositions (Lib1,   LIB1);
      Input[i]->WritePositions (Lib2,   LIB2);
      Input[i]->WritePositions (PubLib, PUBLIB);
   } // for

   for (i = 0; i < Libs; i++)
   {
      Input[i]->WriteFunctions (Lib1,   LIB1,   Input);
      Input[i]->WriteFunctions (Lib2,   LIB2,   Input);
      Input[i]->WriteFunctions (PubLib, PUBLIB, Input);
   } // for

   PubLib << "; init code (second copy of the first code)\n"
             "__init_code:\n";

   for (i = 0; i < Libs; i++)
   {
      Input[i]->Second_Pass ();
      Input[i]->WriteFunctions (PubLib, PUBLIB, Input);
   } // for

   PubLib << "__end_code:\n\n"
          << "publib@" << TMP << ":\n"
          << "init_publib:\n"
          << "\tlea\t__init_code(PC),a0\n"
          << "\tlea\t__start(PC),a1\n"
          << "\tmove.w\t#__end_code-__init_code-1,d0\n"
          << "CopyGlobals:\n"
          << "\tmove.b\t(a0)+,(a1)+\n"
          << "\tdbra.w\td0,CopyGlobals\n"
          << "\trts\n";

   Lib1   << "\n\n_library dc.b \"lib1\",0\n\n\tEND";
   Lib2   << "\n\n_library dc.b \"lib2\",0\n\n\tEND";
   PubLib << "\n\n_library dc.b \"publib\",0\n\n\tEND";

   cout << "Statistics: \n";
   L1_Funcs = 0; L2_Funcs = 0; Publib_Funcs = 0; TotalFuncs = 0;
   for (i = 0; i < Libs; i++)
   {
      TotalFuncs   += Input[i]->GetFunctions ();
      L1_Funcs     += Input[i]->MovedFunctions (LIB1);
      L2_Funcs     += Input[i]->MovedFunctions (LIB2);
      Publib_Funcs += Input[i]->MovedFunctions (PUBLIB);
   } // for

   cout << "Total function " << TotalFuncs   << "\n"
        << "Library 1      " << L1_Funcs     << "\n"
        << "Library 2      " << L2_Funcs     << "\n"
        << "Publib         " << Publib_Funcs << "\n";

   cout << "Generating main module (ontic program)\n";
   if (Doors)
   {
      ifstream IPrgm ("ontic.doo");
      ofstream OPrgm ("ontic.asm");
      char IString[256];

      // First copy until
      while (1)
      {
         IPrgm.getline (IString, 256);
         if (strcmp (IString, "_main:")) break;
         OPrgm << IString << "\n";
      } // while

      // Insert library entries
      for (i = 0; i < Libs; i++)
         Input[i]->WritePositions (OPrgm, EVERYFUNCTION);

      OPrgm << "init_publib\tequ\tpublib@" << TMP << "\n";
      // Copy the rest of the init code
      while (!IPrgm.eof ())
      {
         IPrgm.getline (IString, 256);
         OPrgm << IString << "\n";
      } // while

      // Add cmain code
      ifstream I_cmain ("cmain.s");
      while (1)
      {
         I_cmain.getline (IString, 256);
         if (strstr (IString, "END") != NULL) break;
         OPrgm << IString << "\n";
      } // while

      // end
      OPrgm << "_comment dc.b \"ontic by mxm projects\",0\n\tEND";

   } else {
      ifstream IPrgm ("ontic.far");
      ofstream OPrgm ("ontic.asm");
      char IString[256];

      // First copy until
      while (1)
      {
         IPrgm.getline (IString, 256);
         if (strcmp (IString, "_main:")) break;
         OPrgm << IString << "\n";
      } // while

      // Insert library entries
      for (i = 0; i < Libs; i++)
         Input[i]->WritePositions (OPrgm, EVERYFUNCTION);

      OPrgm << "init_publib\tequ\tpublib@" << TMP << "\n";
      // Copy the rest of the init code
      while (!IPrgm.eof ())
      {
         IPrgm.getline (IString, 256);
         OPrgm << IString << "\n";
      } // while

      // Add cmain code
      ifstream I_cmain ("cmain.s");
      while (1)
      {
         I_cmain.getline (IString, 256);
         if (strstr (IString, "END") != NULL) break;
         OPrgm << IString << "\n";
      } // while

      // end
      OPrgm << "_comment dc.b \"ontic by mxm projects\",0\n\tEND";
   } // if

   cout << "Deleting codetree\n";
   for (i = 0; i < Libs; i++)
      delete Input[i];

   cout << "Done\n";

   return 0;
} // main

